/********************************************************************
Copyright (c) 2001,2002, The Board of Trustees of the University of Illinois.
All Rights Reserved.

SABUL High Performance Data Transfer Protocol

Laboratory for Adavanced Computing (LAC)
National Center for Data Mining (NCDM)
University of Illinois at Chicago
http://www.lac.uic.edu/

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software (SABUL) and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimers.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimers in the documentation
and/or other materials provided with the distribution.

Neither the names of the University of Illinois, LAC/NCDM, nor the names
of its contributors may be used to endorse or promote products derived
from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE CONTIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
********************************************************************/


/********************************************************************
sabul.h
The protocol common data structure and operation definitions

Author: Yunhong Gu [gu@lac.uic.edu]
Last Update: Jan. 07, 2003
********************************************************************/



#ifndef _SABUL_H_
#define _SABUL_H_

using namespace std;

#include <pthread.h>
#include <sys/time.h>
#include <fstream>
#include "channel.h"
#include "packet.h"


struct SabulOption
{
   int m_iDataPort;		// Data Channel Port
   int m_iRecvFlagSize;		// Flow control window size
   int m_iInitRate;		// Requested Initial Sending Rate
   int m_iLLimit;		// Lower limit of sending rate
   int m_iULimit;		// Upper limit of sending rate

   int m_iPktSize;		// for back compatibility only
   int m_iVerbose;		// for back compatibility only
   long m_lSendBufSize;		// for back compatibility only
   long m_lRecvBufSize;		// for back compatibility only
   int m_iPayloadSize;		// for back compatibility only
   int m_iMaxLossLength;	// for back compatibility only
};


// The alias names of SABUL options
// SBL_BLOCK <--> m_bBlockable
// SBL_VERBOSE <--> m_bVerbose
// The others are not used in the current version
enum SblOpt 
{
   SBL_BLOCK = 1, 		// If this value is 1, SABUL enables blocking sending and the sent data buffer
   				// is NOT auto-released. If it is 0, the sending is non-blocking and data will
   				// will be released after it is sent.

   SBL_DPORT, 			// NOT available
   SBL_CPORT, 			// NOT available
   SBL_SNDBUF, 			// NOT available
   SBL_RCVBUF, 			// NOT available
   SBL_PKTSIZE, 		// NOT available
   SBL_RCVFLAGSIZE, 		// NOT available

   SBL_VERBOSE			// Output the debug information to screen (1) or a log file (0).
};

class CSabul
{
public:
   CSabul();
   virtual ~CSabul();
   
   // public API.
   virtual int open(const int& port = -1) {return m_iDefaultPort;}
   virtual int open(const char* ip, const int& port = -1) {return m_iDefaultPort;}
   virtual void close() = 0;

   virtual bool send(char* data, const long& len) const {return false;}
   virtual long recv(char* data, const long& len) const {return 0;}

   virtual bool sendfile(const int& fd, const int& offset, const int& size) const {return false;}
   virtual long recvfile(const int& fd, const int& offset, const int& size) const {return 0;}

   virtual int setOpt(SblOpt optName, const void* optval, const int& optlen);
   virtual int getOpt(SblOpt optName, void* optval, int& optlen);

protected:
   // Timing facilities.

   // rdtsc reads the CPU clock cycle on IA-32 arch.
   // it is the same name as IA-32 instruction of RDTSC.
   void rdtsc(unsigned long long int &x);

   unsigned long long int getCPUFrequency();
   void sleep(unsigned long long int interval);
   void sleepto(unsigned long long int nexttime);

protected:
   // The data and control channels.
   CChannel* m_pDataChannel; 		// DATA channel
   CChannel* m_pCtrlChannel; 		// Control channel

   pthread_t* m_pDCThread; 		// sending/receiving thread

   volatile bool m_bClosing; 		// status: if sabul is closing now.

   unsigned long long int m_ullCPUFrequency; // CPU frequency of the host
 
   int m_iRTT; 				// RTT

   int m_iPktSize; 			// regular packet size
   int m_iPayloadSize; 			// regular payload size
   int m_iMaxLossLength; 		// maximum number of packets can be carried in the ERR message

   int m_iDataPort; 			// Data port, only for DEBUG info.
   int m_iCtrlPort; 			// Control port

   bool m_bBlockable; 			// option: blocking or non-blocking
 
   SabulOption m_RemoteOpt; 		// peer side options.

   // The four kinds of control packets.
   enum PktType {SC_ACK = 1, SC_ERR, SC_EXP, SC_SYN};

   const int m_iDefaultPktSize;		// Default packet size
   const int m_iDefaultPort; 		// Default port

   const long m_lMaxSeqNo; 		// The maximum sequence number
   const long m_lSeqNoTH; 		// For judging if the smaller is prior or after a large one.

   bool m_bConnection; 			// status: If sabul is currently connected.

   long m_lSendBufSize; 		// UDP sending buffer size
   long m_lRecvBufSize; 		// UDP receiving buffer size

   int m_bVerbose; 			// option: output DEBUG info to screen or disk log
   ofstream m_Log; 			// The log file
};



//
// This is a facility class to provide MUTEX lock and unlock
// to a whole class method/function.
//

class CGuard
{
public:
   CGuard(pthread_mutex_t& lock);
   ~CGuard();

private:
   pthread_mutex_t& m_Mutex;		// alias: mutex lock
   int m_iLocked;			// lock status
};

#endif
