/********************************************************************
Copyright (c) 2001,2002, The Board of Trustees of the University of Illinois.
All Rights Reserved.

SABUL High Performance Data Transfer Protocol

Laboratory for Adavanced Computing (LAC)
National Center for Data Mining (NCDM)
University of Illinois at Chicago
http://www.lac.uic.edu/

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software (SABUL) and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimers.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimers in the documentation
and/or other materials provided with the distribution.

Neither the names of the University of Illinois, LAC/NCDM, nor the names
of its contributors may be used to endorse or promote products derived
from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE CONTIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
********************************************************************/


/********************************************************************
sender.cpp
The sender side algorithms, rate control,  and memory management

Author: Yunhong Gu [gu@lac.uic.edu]
Last Update: Mar. 16, 2003
********************************************************************/


#include <pthread.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/mman.h>
#include <iostream>
#include "sender.h"


//
// The constants here MUST NOT be changed.
//
CSabulSender::CSabulSender():
m_iEXPInterval(1000000),
m_dLossRateLimit(0.001),
m_dWeight(0.1)
{
   m_pBuffer = NULL;
   m_pLossList = NULL;

   // mutex and condition variables initilization
   pthread_mutex_init(&m_SendDataLock, NULL);
   pthread_cond_init(&m_SendDataCond, NULL);
   pthread_mutex_init(&m_SendBlockLock, NULL);
   pthread_cond_init(&m_SendBlockCond, NULL);

   // The initial value of the sending rate.
   m_dInterval = 10.0;
   m_dLowerLimit = 1000.0;
   m_dUpperLimit = 10.0;

   // By default it is non-blocking, not output to screen.
   m_bBlockable = false;
   m_bVerbose = false;
}

CSabulSender::~CSabulSender()
{
   if (m_pBuffer)
      delete m_pBuffer;
   if (m_pLossList)
      delete m_pLossList;
}

//
// "open" method initilize the sender and bind it to the "port" number.
// Returned value is the real port number it is bound to.
//

int CSabulSender::open(const int& port)
{
   m_bClosing = false;
   m_lLastAck = 0;
   m_ullInterval = (unsigned long long)(m_dInterval * m_ullCPUFrequency);
   m_lLocalSend = 0;
   m_lLastSYNSeqNo = 0;
   m_lCurrSeqNo = 0;
   m_dHistoryRate = 0.0;
   m_iERRCount = 0;

   m_iDataPort = m_iCtrlPort = port;

   try
   {
      m_pDataChannel = new CUdpChannel(m_iDataPort);
      m_pCtrlChannel = new CTcpChannel(m_iCtrlPort);
   }
   catch(...)
   {

      if (!m_bVerbose)
         m_Log << "Channel Initlization Failed." << endl;
      else
         cout << "Channel Initlization Failed." << endl;

      // Let the application to handle the exception
      throw;
   }


   if (!m_bVerbose)
      m_Log << "Data Channel is ready at port: " << m_iDataPort << endl;
   else
      cout<< "Data Channel is ready at port: " << m_iDataPort << endl;
   if (!m_bVerbose)
      m_Log << "Control Channel is ready at port: " << m_iCtrlPort << endl;
   else
      cout<< "Control Channel is ready at port: " << m_iCtrlPort << endl;


   // Return the actual control channel port.
   return m_iCtrlPort;
}

//
// Listening to a receiver to connect to and exchange information.
//

void CSabulSender::listen(timeval* timeo)
{
   try
   {
      sockaddr_in remoteaddr = m_pCtrlChannel->connect(timeo);
      // Get the receiver's IP
      char* ip = inet_ntoa(remoteaddr.sin_addr);

      // Exchange information
      m_pCtrlChannel->recv((char *)&m_RemoteOpt, sizeof(SabulOption));
      m_pCtrlChannel->send((char *)&m_RemoteOpt, sizeof(SabulOption));

      m_pDataChannel->connect(ip, m_RemoteOpt.m_iDataPort);
   }
   catch(...)
   {

      if (!m_bVerbose)
         m_Log << "Channel Connection Failed." << endl;
      else
         cout << "Channel Connection Failed." << endl;

      // Let the application to handle the exception
      throw;
   }

   // Set the sender window size
   m_iSendFlagSize = m_RemoteOpt.m_iRecvFlagSize;
   m_iWarningLine = m_iSendFlagSize;

   setRate(m_RemoteOpt.m_iInitRate, m_RemoteOpt.m_iLLimit, m_RemoteOpt.m_iULimit);

   m_pBuffer = new CBuffer(m_lSendBufSize);
   m_pLossList = new CLossList;

   // Now it is connected.
   m_bConnection = true;

   pthread_create(m_pDCThread, NULL, CSabulSender::dcHandler, this);
}

//
// Close the connection.
//

void CSabulSender::close()
{
   m_bClosing = true;

   //Signal the sender thread in case it is waiting for the data
   pthread_cond_signal(&m_SendDataCond);

   // Waiting for the thread to exit
   pthread_join(*m_pDCThread, NULL);

   // destroy the channels
   m_pDataChannel->disconnect();
   m_pCtrlChannel->disconnect();
   delete m_pDataChannel;
   delete m_pCtrlChannel;

   delete m_pBuffer;
   delete m_pLossList;
   m_pBuffer = NULL;
   m_pLossList = NULL;
}

//
// The main SABUL sending algorithm can be found in this private method.
// dcHandler() is used as a thread handle that keeps on sending data and receiving feedback.
//

void* CSabulSender::dcHandler(void* sender)
{
   CSabulSender* self = static_cast<CSabulSender *>(sender);

   CPacketVector sdpkt(self->m_iPktSize);
   int payload;
   long offset;

   CPktSblCtrl scpkt(self->m_iPktSize);

   unsigned long long int entertime;
   unsigned long long int nextexptime;
   self->rdtsc(nextexptime);

   unsigned long long int ullexpint = self->m_iEXPInterval * self->m_ullCPUFrequency;

   nextexptime += ullexpint;

   // Record the last SYN time, for calculation of sending rate
   gettimeofday(&(self->m_LastSYNTime), 0);

   while (!self->m_bClosing)
   {
      self->rdtsc(entertime);

      //
      // if the sender hasn't received any feedback from the receiver, 
      // the EXP timer expires and generates an EXP event.
      //

      if (entertime > nextexptime)
      {
         self->processFeedback(SC_EXP);
         nextexptime = entertime + ullexpint;
      }

      //
      // if there is no data to send, wait here...
      //

      if (0 == self->m_pBuffer->getCurrBufSize())
      {
         pthread_mutex_lock(&(self->m_SendDataLock));
         while ((0 == self->m_pBuffer->getCurrBufSize()) && (!self->m_bClosing))
            pthread_cond_wait(&(self->m_SendDataCond), &(self->m_SendDataLock));
         pthread_mutex_unlock(&(self->m_SendDataLock));

         self->rdtsc(nextexptime);
         nextexptime += ullexpint;
      }

      //
      // Poll the control channel to see if there is any feedback...
      //

      *(CTcpChannel *)(self->m_pCtrlChannel) >> scpkt;
      if (scpkt.getLength() > 0)
      {
         //
         // Process them if there are any...
         //
         self->processFeedback(PktType(scpkt.m_lPktType), scpkt.m_lAttr, scpkt.m_plData);
         //
         // Reset the EXP timer...
         //
         if (scpkt.m_lPktType != SC_SYN)
         {
            self->rdtsc(nextexptime);
            nextexptime += ullexpint;
         }
      }
      else if (scpkt.getLength() < 0)
      {
         //
         // Something wrong with the connection...broken!
         //
         self->m_bConnection = false;
         break;
      }

      //
      // Now sending, check the loss list first...
      //
      if ((sdpkt.m_lSeqNo = self->m_pLossList->getLostSeq()) >= 0)
      {
         if ((sdpkt.m_lSeqNo >= self->m_lLastAck) && (sdpkt.m_lSeqNo < self->m_lLastAck + self->m_lSeqNoTH))
            offset = (sdpkt.m_lSeqNo - self->m_lLastAck) * self->m_iPayloadSize;
         else if (sdpkt.m_lSeqNo < self->m_lLastAck - self->m_lSeqNoTH)
            offset = (sdpkt.m_lSeqNo + self->m_lMaxSeqNo - self->m_lLastAck) * self->m_iPayloadSize;
         else
            continue;

         if ((payload = self->m_pBuffer->readData(&(sdpkt.m_pcData), offset, self->m_iPayloadSize)) == 0)
            continue;

         self->m_lLocalSend ++;
      }
      //
      // If there is no loss, send a new packet.
      //
      else
      {
         if (self->m_lCurrSeqNo - self->m_lLastAck >= self->m_iSendFlagSize)
            continue;

         if ((payload = self->m_pBuffer->readData(&(sdpkt.m_pcData), self->m_iPayloadSize)) == 0)
            continue;

         sdpkt.m_lSeqNo = self->m_lCurrSeqNo ++;
         self->m_lCurrSeqNo %= self->m_lMaxSeqNo;
      }
 
      sdpkt.setLength(payload);

      //
      // Convert to network order...
      //
      sdpkt.m_lSeqNo = htonl(sdpkt.m_lSeqNo);

      //
      // Now sending...
      //
      *(CUdpChannel *)(self->m_pDataChannel) << sdpkt;

      //
      // Waiting for the inter-packet time.
      //
      self->sleepto(entertime + self->m_ullInterval);
   }

   return NULL;
}

void CSabulSender::processFeedback(const PktType type, const long attr, const long* data)
{
   switch (type)
   {
   case SC_ACK: // Positive acknowledgement
      // Update the sender buffer
      if (attr >= m_lLastAck)
         m_pBuffer->ackData((attr - m_lLastAck) * m_iPayloadSize, m_iPayloadSize);
      else
         m_pBuffer->ackData((attr - m_lLastAck + m_lMaxSeqNo) * m_iPayloadSize, m_iPayloadSize);
 
      // Update the last ACK variable
      m_lLastAck = attr;

      // Remove all the loss in the loss list prior to the last ACK
      m_pLossList->remove(m_lLastAck);

      // Signal the blocked "send" call to return
      pthread_mutex_lock(&m_SendBlockLock);
      if ((m_bBlockable) && (0 == m_pBuffer->getCurrBufSize()))
         pthread_cond_signal(&m_SendBlockCond);
      pthread_mutex_unlock(&m_SendBlockLock);

      break;

   case SC_SYN: // Calcualte loss rate and do rate control here
      // The total number of sent packets equal to the sum of the retransmission and the new sent packets 
      m_lLocalSend += m_lCurrSeqNo - m_lLastSYNSeqNo;

      // A loop of the sequence number
      if (m_lCurrSeqNo < m_lLastSYNSeqNo)
         m_lLocalSend += m_lMaxSeqNo;

      // Total sent cannot be less than total loss 
      if (m_lLocalSend < m_lLocalERR)
         m_lLocalSend = m_lLocalERR;

      if (m_lLocalSend != 0)
      {
         //Print loss and sending rate.
         if (!m_bVerbose)
            m_Log << "loss rate: "<< double(m_lLocalERR) / double(m_lLocalSend) * 100 << "%" << "\t\t";
         else
         {
            cout<< "loss rate: ";
            cout.width(10);
            cout<< double(m_lLocalERR) / double(m_lLocalSend) * 100 << "%" << "\t\t";
         }
 
         rateControl(double(m_lLocalERR) / double(m_lLocalSend));

         timeval currtime;
         gettimeofday(&currtime, 0);

         if (!m_bVerbose)
            m_Log << "sending rate: "<< double(m_lLocalSend) * 1500.0 * 8.0 / ((currtime.tv_sec - m_LastSYNTime.tv_sec) * 1000000 + (currtime.tv_usec - m_LastSYNTime.tv_usec)) << "Mbps" << endl;
         else
            cout<< "sending rate: "<< double(m_lLocalSend) * 1500.0 * 8.0 / ((currtime.tv_sec - m_LastSYNTime.tv_sec) * 1000000 + (currtime.tv_usec - m_LastSYNTime.tv_usec)) << "Mbps" << endl;

         m_ullInterval = (unsigned long long)(m_dInterval * m_ullCPUFrequency);
      }

      // Record the last SYN time.
      gettimeofday(&m_LastSYNTime, 0);

      m_lLastSYNSeqNo = m_lCurrSeqNo;
      m_lLocalSend = 0;
      m_lLocalERR = 0;
      m_iERRCount = 0;

      break;

   case SC_ERR: // Loss, insert the loss seq. no. into the loss list.
      // Increase the sending interval if there is too much loss
      if (10 == ++ m_iERRCount)
      {
         m_dInterval += 2.0;
         m_ullInterval += 2 * m_ullCPUFrequency;
         m_iERRCount = -500;
      }

      // Insert the loss into the loss list
      for (int i = attr - 1; (i >= 0) && (m_pLossList->insert(data[i])); i --) {}

      m_lLocalERR += attr;

      break;

   case SC_EXP:
      // Insert all the packets into the loss list since last ACK sequence number.
      for (int i = m_lCurrSeqNo - m_lLastAck - 1; (i >= 0) && (m_pLossList->insert(i + m_lLastAck)); i --) {}
 
      break;

   default:
      break;
   }
}

//
// Part of the rate control, the other part is in the processFeedback() SC_ERR section.
//

void CSabulSender::rateControl(const double& currlossrate)
{
   // calculate loss rate
   m_dHistoryRate = m_dHistoryRate * m_dWeight + currlossrate * (1 - m_dWeight);

   if (m_dHistoryRate > m_dLossRateLimit)
      m_dInterval = m_dInterval * (1. + 0.1 * (m_dHistoryRate - m_dLossRateLimit)) + 0.5;
   else if (m_dHistoryRate < m_dLossRateLimit)
      m_dInterval = m_dInterval * (1. + 10.0 * (m_dHistoryRate - m_dLossRateLimit));
   else
      m_dInterval += 0.1;

   // limit the sending rate
   if (m_dInterval > m_dLowerLimit)
         m_dInterval = m_dLowerLimit;
   else if (m_dInterval < m_dUpperLimit)
         m_dInterval = m_dUpperLimit;
}

//
// API send
// 

bool CSabulSender::send(const char* data, const long& len)
{
   if (!m_bConnection)
      throw int(-1);

   if (len > m_pBuffer->getSizeLimit())
      throw int(-2);

   bool ret;

   // Insert the application buffer into the sender buffer linked list
   pthread_mutex_lock(&m_SendDataLock);
   ret = m_pBuffer->addBuffer(data, len, m_bBlockable);
   pthread_mutex_unlock(&m_SendDataLock);

   // Inform the sending thread that new data comes
   if (ret)
   {
      pthread_cond_signal(&m_SendDataCond);

      // Block here if it is in blocking mode
      pthread_mutex_lock(&m_SendBlockLock);
      while ((m_bBlockable) && (0 != m_pBuffer->getCurrBufSize()))
         pthread_cond_wait(&m_SendBlockCond, &m_SendBlockLock);
      pthread_mutex_unlock(&m_SendBlockLock);
   }

   return ret;
}

//
// API sendfile
//

bool CSabulSender::sendfile(const int& fd, const int& offset, const int& size)
{
   if (!m_bConnection)
      throw int(-1);

   // sendfile is always blocking.
   bool block = m_bBlockable;
   m_bBlockable = false;

   char* tempbuf;

   // send 7.34MB each time
   long unitsize = 7340000;
   int count = 1;

   while (unitsize * count <= size)
   {
      tempbuf = new char[unitsize];
      read(fd, tempbuf, unitsize);

      pthread_mutex_lock(&m_SendDataLock);
      while (!m_pBuffer->addBuffer(tempbuf, unitsize))
         usleep(10);
      pthread_mutex_unlock(&m_SendDataLock);

      // signal sending thread
      pthread_cond_signal(&m_SendDataCond);

      count ++;
   }
   if (size - unitsize * (count - 1) > 0)
   {
      tempbuf = new char[size - unitsize * (count - 1)];
      read(fd, tempbuf, size - unitsize * (count - 1));

      pthread_mutex_lock(&m_SendDataLock);
      while (!m_pBuffer->addBuffer(tempbuf, size - unitsize * (count - 1)))
         usleep(10);
      pthread_mutex_unlock(&m_SendDataLock);

      // singal sending thread
      pthread_cond_signal(&m_SendDataCond);

      count ++;
   }

   while (getCurrBufSize())
      usleep(10);

//
// The following mmap solution has unknown mistakes...
// So it is NOT used any more.
//

/*
   size_t pagesize = getpagesize(); //4096
   long mapsize = pagesize * (m_iPayloadSize >> 2);
   char* filemap;
   int count = 1;

   while (mapsize * count <= size)
   {
      filemap = (char *)mmap(0, mapsize, PROT_READ, MAP_SHARED, fd, (count - 1) * mapsize);

      pthread_mutex_lock(&m_SendDataLock);
      while (!m_pBuffer->addBuffer(filemap, mapsize, true))
         usleep(10);
      pthread_mutex_unlock(&m_SendDataLock);

      pthread_cond_signal(&m_SendDataCond);

      count ++;
   }
   if (size - mapsize * (count - 1) > 0)
   {
      filemap = (char *)mmap(0, size - mapsize * (count - 1), PROT_READ, MAP_SHARED, fd, (count - 1) * mapsize);

      pthread_mutex_lock(&m_SendDataLock);
      while (!m_pBuffer->addBuffer(filemap, size - mapsize * (count - 1), true))
         usleep(10);
      pthread_mutex_unlock(&m_SendDataLock);

      pthread_cond_signal(&m_SendDataCond);
   }
*/

   m_bBlockable = block;

   return true;
}

long CSabulSender::getCurrBufSize() const
{
   // not allowed before connection is set up
   if (!m_bConnection)
      throw int(-1);

   return m_pBuffer->getCurrBufSize();
}

void CSabulSender::setRate(const int& initrate, const int& l_limit, const int& u_limit) //in Mbps
{
   // convert sending rate to interpacket time

   m_dInterval = (1500.0 * 8.0) / initrate;

   if (l_limit > 0)
      m_dLowerLimit = (1500.0 * 8.0) / l_limit;
   else
      m_dLowerLimit = 1000.0;

   m_dUpperLimit = (1500.0 * 8.0) / u_limit;
}

//
// The buffer management codes, see reference in sender.h
//

// Initialize the sender buffer
CSabulSender::CBuffer::CBuffer():
m_pBlock(NULL),
m_pLastBlock(NULL),
m_pCurrSendBlk(NULL),
m_pCurrAckBlk(NULL),
m_lCurrBufSize(0),
m_lSizeLimit(40960000)
{
   pthread_mutex_init(&m_BufLock, NULL);
}

CSabulSender::CBuffer::CBuffer(const long& bufsize):
m_pBlock(NULL),
m_pLastBlock(NULL),
m_pCurrSendBlk(NULL),
m_pCurrAckBlk(NULL),
m_lCurrBufSize(0),
m_lSizeLimit(bufsize)
{
   pthread_mutex_init(&m_BufLock, NULL);
}

CSabulSender::CBuffer::~CBuffer()
{
   Block* pb = m_pBlock;
  
   // delete all unsent block
   while (NULL != m_pBlock)
   {
      pb = pb->m_next;
      if (m_pBlock->m_bMap)
         munmap(m_pBlock->m_pcData, m_pBlock->m_lLength);
      else
         delete [] m_pBlock->m_pcData;
      delete m_pBlock;

      m_pBlock = pb;
   }
}

bool CSabulSender::CBuffer::addBuffer(const char* data, const long& len, const bool map)
{
   // locking...
   CGuard bufferguard(m_BufLock);

   // cannot add more buffer to the list
   if (m_lCurrBufSize + len > m_lSizeLimit)
      return false;

   // If the buffer is empty, we now have the first block.
   if (NULL == m_pBlock)
   {
      m_pBlock = new Block;
      m_pBlock->m_pcData = const_cast<char *>(data);
      m_pBlock->m_lLength = len;
      m_pBlock->m_bMap = map;
      m_pBlock->m_next = NULL;
      m_pLastBlock = m_pBlock;
      m_pCurrSendBlk = m_pBlock;
      m_lCurrSendPnt = 0;
      m_pCurrAckBlk = m_pBlock;
      m_lCurrAckPnt = 0;
   }
   // Otherwise, insert the block at the tail of the list.
   else
   {
      m_pLastBlock->m_next = new Block;
      m_pLastBlock = m_pLastBlock->m_next;
      m_pLastBlock->m_pcData = const_cast<char *>(data);
      m_pLastBlock->m_lLength = len;
      m_pLastBlock->m_bMap = map;
      m_pLastBlock->m_next = NULL;
      if (NULL == m_pCurrSendBlk)
         m_pCurrSendBlk = m_pLastBlock;
   }

   // Update the current buffer size
   m_lCurrBufSize += len;

   return true;
}

long CSabulSender::CBuffer::readData(char** data, const long& len)
{
   if (NULL == m_pCurrSendBlk)
      return 0;

   // Read data from m_lCurrSendPnt with a regular data size
   if (m_lCurrSendPnt + len < m_pCurrSendBlk->m_lLength)
   {
      *data = m_pCurrSendBlk->m_pcData + m_lCurrSendPnt;
      m_lCurrSendPnt += len;
      return len;
   }

   // It there is not enough data in the current block
   // read all the rest data to pack a smaller irregular packet
   // The receiver will indentify it.
   long readlen = m_pCurrSendBlk->m_lLength - m_lCurrSendPnt;
   *data = m_pCurrSendBlk->m_pcData + m_lCurrSendPnt;
 
   // Move to next block
   m_pCurrSendBlk = m_pCurrSendBlk->m_next;
   m_lCurrSendPnt = 0;     
    
   return readlen;
}

long CSabulSender::CBuffer::readData(char** data, const long offset, const long& len)
{
   CGuard bufferguard(m_BufLock);

   Block* p = m_pCurrAckBlk;
   if (NULL == p)
      return 0;
   long loffset = offset + m_lCurrAckPnt;

   // Look for the data position
   while (p->m_lLength <= loffset)
   {
      loffset -= p->m_lLength;
      loffset -= len - ((0 == p->m_lLength % len) ? len : (p->m_lLength % len));
      p = p->m_next;
      if (NULL == p)
         return 0;
   }

   // Read regular sized data
   if (loffset + len <= p->m_lLength)
   {
      *data = p->m_pcData + loffset;
      return len;
   }

   // Read irregular sized data
   *data = p->m_pcData + loffset;
   return p->m_lLength - loffset;
}

void CSabulSender::CBuffer::ackData(const long& len, const int& payloadsize)
{
   CGuard bufferguard(m_BufLock);

   m_lCurrAckPnt += len;

   // if a block has been totally sent and acknowledged
   while (m_lCurrAckPnt >= m_pCurrAckBlk->m_lLength)
   {
      m_lCurrAckPnt -= m_pCurrAckBlk->m_lLength;
      if (0 != m_pCurrAckBlk->m_lLength % payloadsize)
         m_lCurrAckPnt -= payloadsize - m_pCurrAckBlk->m_lLength % payloadsize;

      m_lCurrBufSize -= m_pCurrAckBlk->m_lLength;
      m_pCurrAckBlk = m_pCurrAckBlk->m_next;

      // Delete the block in SABUL or not according to the map flag
      //
      // Should be further refined here
      //
      if (m_pBlock->m_bMap)
         munmap(m_pBlock->m_pcData, m_pBlock->m_lLength);
      else
         delete [] m_pBlock->m_pcData;

      // release the block sturcture
      delete m_pBlock;

      m_pBlock = m_pCurrAckBlk;

      // if no block left, return
      if (NULL == m_pBlock)
         break;
   }
}

long CSabulSender::CBuffer::getCurrBufSize() const
{
   return m_lCurrBufSize;
}

long CSabulSender::CBuffer::getSizeLimit() const
{
   return m_lSizeLimit;
}

//
// The loss list management codes, see reference in sender.h
//

// The m_lAttr in the first node is the total number of the loss
CSabulSender::CLossList::CLossList():
m_lAttr(0),
m_next(NULL)
{
}

CSabulSender::CLossList::CLossList(const long& seqno):
m_lAttr(seqno),
m_next(NULL)
{
}

int CSabulSender::CLossList::insert(const long& seqno)
{
   CLossList *p = this;

   // search for the right position to keep the order of the list
   while ((NULL != p->m_next) && (((p->m_next->m_lAttr < seqno) && (seqno - p->m_next->m_lAttr < (1 << 20))) || (p->m_next->m_lAttr - seqno > (1 << 20))))
      p = p->m_next;

   if (NULL == p->m_next)
   {
      // insert a node at tail
      p->m_next = new CLossList(seqno);
      m_lAttr ++;
      return 1;
   }

   if (seqno == p->m_next->m_lAttr)
      return 0;

   // insert a node in the middle of the list
   CLossList *q = new CLossList(seqno);
   q->m_next = p->m_next;
   p->m_next = q;
   m_lAttr ++;

   return 1;
}

void CSabulSender::CLossList::remove(const long& seqno)
{
   CLossList *p = this->m_next;

   // remove all the nodes prior to and include the one with the "seqno".
   while ((NULL != p) && (((p->m_lAttr <= seqno) && (seqno - p->m_lAttr < (1 << 20))) || (p->m_lAttr - seqno > (1 << 20))))
   {
      // remove a node
      m_next = p->m_next;
      delete p;
      p = m_next;

      // decrease the total number of loss
      m_lAttr --;
   }
}

int CSabulSender::CLossList::getLossLength() const
{
   return m_lAttr;
}

long CSabulSender::CLossList::getLostSeq()
{
   // empty list, return -1
   if (NULL == m_next)
      return -1;

   // remove a node
   long seqno = m_next->m_lAttr;
   CLossList *p = m_next;
   m_next = p->m_next;
   delete p;

   // decrease the total number of loss
   m_lAttr --;

   return seqno;
}

//
