/* cxextr.c -- Extract two 2-bit sample streams from n-bit files,
               output two 4-bit samples (one byte) per each input word . */

/*
 * $Log$
 *
 */

/* Copyright (C) 2001--2004 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>  /* for open()ing */
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>  /* for read()ing */
#include <assert.h>


typedef unsigned char tByte, *ptByte;


ssize_t
sureRead(int filedes, ptByte memblk, size_t blksize)
{
  size_t nread;

  /* Read a block; if not enough, read more. */
  assert( (nread = read (filedes, memblk, blksize)) >= 0 );
  if (nread <= 0)
    goto lReturn;
  while (nread < blksize) {
    size_t nr;

    /* Not one full block in buffer, read more. */
    assert( (nr = read (filedes, memblk+nread, blksize-nread)) >= 0 );
    if (nr <= 0)
      goto lReturn;
    nread += nr;
  }  /* while not at least one full block in buffer to return */
 lReturn:
  return (nread);
}  /* sureRead */

ssize_t
sureWrite(int filedes, ptByte memblk, size_t blksize)
{
  size_t nwritten;

  /* Write a block; if couldn't write all, write again more. */
  assert( (nwritten = write (filedes, memblk, blksize)) >= 0 );
  while (nwritten < blksize) {
    size_t nw;

    /* Not everything in buffer written, write more. */
    assert( (nw = write (filedes, memblk+nwritten, blksize-nwritten)) >= 0 );
    nwritten += nw;
  }  /* while not everything in buffer written */
  return (nwritten);
}  /* sureWrite */


int
main(int argc, char *argv[])
{
  off_t firstseek;
  int blksize;
  int totalblks;
  int wordbits;
  int wordbytes;
  int wordsPerRdBlk;
  int whichbitlsn;  /* the bit position number of LSB of output LSNibble */
  int whichbitlsnbyte;
  int whichbitlsnmask;
  int whichbitlsnshift;
  int whichbitmsn;  /* the bit position number of LSB of output MSNibble */
  int whichbitmsnbyte;
  int whichbitmsnmask;
  int whichbitmsnshift;
  tByte sm[4];  /* sample mapping for the four possible 2-bit values */
  double lsnss[5];  /* LSNibble sampler stats, [4]==total */
  double msnss[5];  /* MSNibble sampler stats */
  int blsnss[5];  /* intra-block LSNibble sampler stats, [4]==total */
  int bmsnss[5];  /* intra-block MSNibble sampler stats */
  ptByte rdBlk;
  ptByte wrBlk;
  int i;
  int b;

  if (argc < 6) {
    fprintf(stderr, "%s: needs at least blocksize wordbits whichbitlsn whichbitmsn bitmapping\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  /* Command line arguments. */
  firstseek = /** atoi(argv[?]) **/ 0;  /* disable for now */
  blksize = atoi(argv[1]);
  totalblks = /** atoi(argv[?]) **/ 0;  /* unlimited, until EOF */
  wordbits = atoi(argv[2]);
  assert( ((wordbits % 8) == 0) );
  wordbytes = wordbits / 8;
  assert( ((blksize % wordbytes) == 0) );
  wordsPerRdBlk = blksize / wordbytes;
  assert( (rdBlk = malloc(blksize)) != NULL );
  assert( (wrBlk = malloc(wordsPerRdBlk)) != NULL );

  whichbitlsn = atoi(argv[3]);
  assert( ((whichbitlsn >= 0) && (whichbitlsn <= wordbits)) );
  whichbitlsnbyte = whichbitlsn / 8;
  whichbitlsnshift = whichbitlsn % 8;
  whichbitlsnmask = (0x03 << whichbitlsnshift);

  whichbitmsn = atoi(argv[4]);
  assert( ((whichbitmsn >= 0) && (whichbitmsn <= wordbits)) );
  whichbitmsnbyte = whichbitmsn / 8;
  whichbitmsnshift = whichbitmsn % 8;
  whichbitmsnmask = (0x03 << whichbitmsnshift);

  /* 2-bit sample coding mapping, "0123 -> 0123". */
  /* (Output code can be hex, 0..f for 4-bit output codes.) */
  assert( strlen(argv[5]) == 4 );
  {
    char hexstring[] = "0x0";

    hexstring[2] = argv[5][0]; sm[0] = strtol(hexstring, (char **)NULL, 0);
    hexstring[2] = argv[5][1]; sm[1] = strtol(hexstring, (char **)NULL, 0);
    hexstring[2] = argv[5][2]; sm[2] = strtol(hexstring, (char **)NULL, 0);
    hexstring[2] = argv[5][3]; sm[3] = strtol(hexstring, (char **)NULL, 0);
  }

  fprintf(stderr, "%s: lsn: extracting 2bits masked by 0x%02X of byte %d of each %d-bit word\n", argv[0], whichbitlsnmask, whichbitlsnbyte, wordbits);
  fprintf(stderr, "%s: msn: extracting 2bits masked by 0x%02X of byte %d of each %d-bit word\n", argv[0], whichbitmsnmask, whichbitmsnbyte, wordbits);

  /* Skip some if needed.  (This for disk files for now.) */
  if (firstseek > 0) {
    assert( (lseek(STDIN_FILENO, firstseek, SEEK_SET) == firstseek) );
    fprintf(stderr, "%s: skipped to %ld bytes\n", argv[0], firstseek);
  }

  /* Clear total sampler statistics. */
  lsnss[0] = lsnss[1] = lsnss[2] = lsnss[3] = lsnss[4] = 0.0;
  msnss[0] = msnss[1] = msnss[2] = msnss[3] = msnss[4] = 0.0;

  /* Read all blocks requested. */
  for (i=0; /** i<totalblks */ ; i++) {
    ssize_t thisRead;

    thisRead = sureRead(STDIN_FILENO, rdBlk, blksize);
    if (thisRead < blksize) {
      /* Assume EOF, skip the last partial block for now. */
      fprintf(stderr, "%s: EOF at block %d, partial read bytes = %d\n", argv[0], i, thisRead);
      break;  /* from for loop */
    }

    /* Clear intra-block sampler statistics. */
    blsnss[0] = blsnss[1] = blsnss[2] = blsnss[3] = blsnss[4] = 0;
    bmsnss[0] = bmsnss[1] = bmsnss[2] = bmsnss[3] = bmsnss[4] = 0;

    /* For each block read, format the output and write it out. */
    for (b=0; b<wordsPerRdBlk; b++) {
      unsigned char s1, s2;

      s1 = (rdBlk[b*wordbytes + whichbitlsnbyte] & whichbitlsnmask) >> whichbitlsnshift;
      s2 = (rdBlk[b*wordbytes + whichbitmsnbyte] & whichbitmsnmask) >> whichbitmsnshift;
      wrBlk[b] = sm[s1] | (sm[s2])<<4;

      /* Update sampler statistics. */
      blsnss[s1]++;
      blsnss[4]++;
      bmsnss[s2]++;
      bmsnss[4]++;
    }  /* for each output byte */

    /* Write out one block of result bytes. */
    sureWrite(STDOUT_FILENO, wrBlk, wordsPerRdBlk);

    /* Update per-block sampler stats. */
    lsnss[0] += (((double)blsnss[0]) / ((double)blsnss[4]));
    lsnss[1] += (((double)blsnss[1]) / ((double)blsnss[4]));
    lsnss[2] += (((double)blsnss[2]) / ((double)blsnss[4]));
    lsnss[3] += (((double)blsnss[3]) / ((double)blsnss[4]));
    lsnss[4] += 1.0;
    msnss[0] += (((double)bmsnss[0]) / ((double)bmsnss[4]));
    msnss[1] += (((double)bmsnss[1]) / ((double)bmsnss[4]));
    msnss[2] += (((double)bmsnss[2]) / ((double)bmsnss[4]));
    msnss[3] += (((double)bmsnss[3]) / ((double)bmsnss[4]));
    msnss[4] += 1.0;
  }  /* for each input block read */

  fprintf(stderr, "%s: lsn sampler stats %d: 0/%f, 1/%f, 2/%f, 3/%f\n", argv[0], whichbitlsn, lsnss[0]/lsnss[4]*2.0, lsnss[1]/lsnss[4]*2.0, lsnss[2]/lsnss[4]*2.0, lsnss[3]/lsnss[4]*2.0);
  fprintf(stderr, "%s: msn sampler stats %d: 0/%f, 1/%f, 2/%f, 3/%f\n", argv[0], whichbitmsn, msnss[0]/msnss[4]*2.0, msnss[1]/msnss[4]*2.0, msnss[2]/msnss[4]*2.0, msnss[3]/msnss[4]*2.0);

  exit (EXIT_SUCCESS);
}  /* main */
