/* exb.c -- Extract bits from files, output bytes per 1/2-bit samples. */

/*
 * $Log$
 *
 */

/* Copyright (C) 2001--2004 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>  /* for open()ing */
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>  /* for read()ing */
#include <assert.h>


typedef unsigned char tByte, *ptByte;


ssize_t
sureRead(int filedes, ptByte memblk, size_t blksize)
{
  size_t nread;

  /* Read a block; if not enough, read more. */
  assert( (nread = read (filedes, memblk, blksize)) >= 0 );
  if (nread <= 0)
    goto lReturn;
  while (nread < blksize) {
    size_t nr;

    /* Not one full block in buffer, read more. */
    assert( (nr = read (filedes, memblk+nread, blksize-nread)) >= 0 );
    if (nr <= 0)
      goto lReturn;
    nread += nr;
  }  /* while not at least one full block in buffer to return */
 lReturn:
  return (nread);
}  /* sureRead */

ssize_t
sureWrite(int filedes, ptByte memblk, size_t blksize)
{
  size_t nwritten;

  /* Write a block; if couldn't write all, write again more. */
  assert( (nwritten = write (filedes, memblk, blksize)) >= 0 );
  while (nwritten < blksize) {
    size_t nw;

    /* Not everything in buffer written, write more. */
    assert( (nw = write (filedes, memblk+nwritten, blksize-nwritten)) >= 0 );
    nwritten += nw;
  }  /* while not everything in buffer written */
  return (nwritten);
}  /* sureWrite */


int
main(int argc, char *argv[])
{
  off_t firstseek;
  int blksize;
  int totalblks;
  int wordbits;
  int wordbytes;
  int wordsPerRdBlk;
  int whichbit;
  int whichbitbyte;
  int whichbitmask;
  tByte zerovalue;
  tByte onevalue;
  ptByte rdBlk;
  ptByte wrBlk;
  int i;
  int b;

  if (argc < 8) {
    fprintf(stderr, "%s: needs at least firstseek blocksize totalblocks wordbits whichbit zerovalue onevalue\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  firstseek = atoi(argv[1]);
  blksize = atoi(argv[2]);
  totalblks = atoi(argv[3]);
  wordbits = atoi(argv[4]);
  assert( ((wordbits % 8) == 0) );
  wordbytes = wordbits / 8;
  assert( ((blksize % wordbytes) == 0) );
  wordsPerRdBlk = blksize / wordbytes;
  assert( (rdBlk = malloc(blksize)) != NULL );
  assert( (wrBlk = malloc(wordsPerRdBlk)) != NULL );
  whichbit = atoi(argv[5]);
  assert( ((whichbit >= 0) && (whichbit <= wordbits)) );
  whichbitbyte = whichbit / 8;
  whichbitmask = (1 << (whichbit % 8));
  zerovalue = atoi(argv[6]);
  onevalue = atoi(argv[7]);
  fprintf(stderr, "%s: extracting bit masked by 0x%02X of byte %d of each %d-bit word\n", argv[0], whichbitmask, whichbitbyte, wordbits);

  /* Skip some if needed.  (This for disk files for now.) */
  if (firstseek > 0) {
    assert( (lseek(STDIN_FILENO, firstseek, SEEK_SET) == firstseek) );
    fprintf(stderr, "%s: skipped to %ld bytes\n", argv[0], firstseek);
  }

  /* Read all blocks requested. */
  for (i=0; i<totalblks; i++) {
    sureRead(STDIN_FILENO, rdBlk, blksize);

    /* For each block read, format the output and write it out. */
    for (b=0; b<wordsPerRdBlk; b++) {
      wrBlk[b] = (rdBlk[b*wordbytes + whichbitbyte] & whichbitmask) ? onevalue : zerovalue;
    }  /* for each output byte */

    sureWrite(STDOUT_FILENO, wrBlk, wordsPerRdBlk);
  }  /* for each input block read */

  exit (EXIT_SUCCESS);
}  /* main */
