/* sf.c -- Mk4/VLBA software formatter. */

/*
 * $Log: sf.c,v $
 * Revision 1.5  2005/01/19 07:19:02  amn
 * Dec-2004 JIVE software formatter changes; vsib.c seek support.
 *
 * Revision 1.2  2004/09/17 15:39:27  amn
 * 2-bit sampe coding mapping; sureRead() and sureWrite() for piping.
 *
 * Revision 1.1  2004/08/02 18:35:39  amn
 * First release of sf.c, software Mk4/VLBA formatter.
 *
 *
 */

/* Copyright (C) 2001--2004 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <sys/types.h>  /* open() */
#include <sys/stat.h>  /* open() */
#include <fcntl.h>  /* open() */

#include <sys/time.h>  /* gettimeofday() */
#include <unistd.h>  /* gettimeofday(), usleep() */

#include <string.h>  /* strstr() */


/* Accepting ISO-8601 std. date/time UTC specs and converting to MJD. */
#include <time.h>  /* mktime(), struct tm */
#include <math.h>  /* floor() */
#include "tstamp.c"
#include "njd.c"


/* Formatter time, in calculatable/incrementable struct, */
/* both in DOY and MJD format for both Mk4 and VLBA timecode. */
typedef struct sFormatterTime {
  int yyyy;  /* full 4-digit year */
  int mm;  /* 01--12 */
  int dd;  /* 01--31 */
  int mjd;  /* modified Julian date */
  int doy;  /* day of year, 1==1-Jan */
  int hh;  /* 00--23 */
  int min;  /* 00--59 */
  int ss;  /* 00--59 xxx: no leap seconds... */
  int dsssss;  /* day seconds, 0--86399 */
  int f800;  /* sub-second frame counter, 0--799 */
  /* at 16Mbps incs by +1, 8Mbps by +2, 4 Mbps by +4...0.5Mbps by +25 */
  int frameIncr;
  double trackRate;  /* just for error/info messages */
} tFormatterTime;

/* Formatter state and settings. */
tFormatterTime ft;
char formatterMode;  /* m/v */
int sm[4];  /* 2-bit sample coding mapping, "samplemap" 'sm' */
/* 'sm' only used in init of cross-point tables, so data type */
/* doesn't matter performance-wise. */
int inBits, outBits;  /* 64/32/16/8 --> 64/32/16/8 */
int fanOutVariant;



typedef unsigned char tByte, *ptByte;


/* Mark IV frame header in byte chunks, without parity bits. */
typedef struct sRawMk4Header {
  /* 64-bit/8-byte aux field */
  tByte auxhs1[2];
  tByte auxhs2[2];
  tByte auxtrk;
  tByte auxsrc;
  tByte auxsts;
  tByte auxsid;
  /* 4-byte sync word */
  tByte sync[4];
  /* 64-bit/8-byte time code (52 bits) plus CRC (12 bits) field */
  tByte timec[5];  /* to full seconds */
  tByte timef[3];  /* sub-second plus CRC-12 */
} tRawMk4Header;

/* The format template is shared and retained between the "any-second" */
/* and "sub-second" formatter functions mk4fmtc() and mk4fmtf(). */
tRawMk4Header mf;

/* VLBA frame header in byte chunks, without parity bits. */
typedef struct sRawVLBAHeader {
  /* 4-byte sync word */
  tByte sync[4];
  /* 64-bit/8-byte time code (48 bits) plus CRC (12 bits) field */
  tByte timec[4];  /* MJD to full seconds */
  tByte timef[2];  /* sub-second */
  tByte crc16[2];  /* plus CRC-16 (2 bytes) */
  /* Here in the middle 22500 bits of non-replacement data + parity. */
  /* 64-bit/8-byte aux field */
  tByte auxsid;
  tByte auxhs[3];  /* recorder#, headstack#, hs position */
  tByte auxdar[3];  /* bit-reversed TG BS CD */
  tByte auxunused;  /* 0x00 */
} tRawVLBAHeader;

/* Format template. */
tRawVLBAHeader vf;

/* Special encodings of sub-millisecond time in time fields. */
const tByte msms[8] = {  /* Mk4, in 4 MSBits of the byte */
  /* 0 == 0.000 => */ 0 << 4,
  /* 1 == 0.125 => */ 1 << 4,
  /* 2 == 0.250 => */ 2 << 4,
  /* 3 == 0.375 => */ 3 << 4,
  /* 4 == 0.500 => */ 5 << 4,
  /* 5 == 0.625 => */ 6 << 4,
  /* 6 == 0.750 => */ 7 << 4,
  /* 7 == 0.875 => */ 8 << 4
};
const tByte vsms[8] = {  /* VLBA */
  /* 0 == 0.000 => */ 0x00,
  /* 1 == 0.125 => */ 0x12,  /* extension, truncated value */
  /* 2 == 0.250 => */ 0x25,
  /* 3 == 0.375 => */ 0x37,  /* extension, truncated value */
  /* 4 == 0.500 => */ 0x50,
  /* 5 == 0.625 => */ 0x62,  /* extension, truncated value */
  /* 6 == 0.750 => */ 0x75,
  /* 7 == 0.875 => */ 0x87,  /* extension, truncated value */
};
/* VLBA only defines msec and 0.1msec codes for multiples of 0.25msec. */
/* That's sufficient for 8Mbps, the max of VLBA formatter. */
/* This "extension by truncated value" gives the same 4-bit numbers for */
/* the 1msec digit as what is used in Mk4. */



/* Data words, 64/32/16/8 bits. */
typedef unsigned long long int tD64, *ptD64;
typedef unsigned int tD32, *ptD32;
typedef unsigned short int tD16, *ptD16;
typedef unsigned char tD8, *ptD8;

/* Generate formatter code for 64--8 bit data word widths. */
#define NM(a) a ## 64
#define BITS 64
#include "sff.c"
#undef BITS
#undef NM

#define NM(a) a ## 32
#define BITS 32
#include "sff.c"
#undef BITS
#undef NM

#define NM(a) a ## 16
#define BITS 16
#include "sff.c"
#undef BITS
#undef NM

#define NM(a) a ## 8
#define BITS 8
#include "sff.c"
#undef BITS
#undef NM


/* xxx: */
#define B64 1

/* Crosspoint switch / fanout bit map tables, 64-bit --> 64-bit. */
tD64 xpt0[256];
tD64 xpt1[256];
tD64 xpt2[256];
tD64 xpt3[256];
#if B64
tD64 xpt4[256];
tD64 xpt5[256];
tD64 xpt6[256];
tD64 xpt7[256];
#endif


void
setupFormatterStartTime(char *argv[])
{
  double fractSec;
  double jd;
  double sr;
  double framesPerSec;

  /* Take start UTC time from command line. */
  assert( convDateTime(argv[1], &(ft.yyyy), &(ft.mm), &(ft.dd),
                       &(ft.hh), &(ft.min), &fractSec) );
  jd = julian_date(ft.yyyy, ft.mm, ft.dd, 0.0);
  ft.mjd = jd - 2400000.5;
  /* xxx: assuming "day of year" starts with 1. */
  ft.doy = jd - julian_date(ft.yyyy, 1, 1, 0.0) + 1;
  ft.ss = floor(fractSec);
  /* xxx: assuming "day seconds" is in range 0--86399. */
  ft.dsssss = (ft.hh*60 + ft.min)*60 + ft.ss;

  /* Get mk4/vlba mode, in-#bits, in-samplerate, out-#bits from cmdline. */
  formatterMode = argv[2][0];  /* m/v */
  assert( (formatterMode == 'm') || (formatterMode == 'v') );

  /* 2-bit sample coding mapping, "0123 -> 0123". */
  assert( strlen(argv[3]) == 4 );
  assert( (argv[3][0] >= '0') && (argv[3][0] <= '3') );
  sm[0] =  argv[3][0] & 0x3;
  assert( (argv[3][1] >= '0') && (argv[3][1] <= '3') );
  sm[1] =  argv[3][1] & 0x3;
  assert( (argv[3][2] >= '0') && (argv[3][2] <= '3') );
  sm[2] =  argv[3][2] & 0x3;
  assert( (argv[3][3] >= '0') && (argv[3][3] <= '3') );
  sm[3] =  argv[3][3] & 0x3;

  inBits = atoi(argv[4]);
  assert( (inBits == 8) || (inBits == 16) || (inBits == 32) || (inBits == 64) );
  sr = atof(argv[5]);
  outBits = atoi(argv[6]);
  fanOutVariant = atoi(argv[7]);
  assert( (outBits == 8) || (outBits == 16) || (outBits == 32) || (outBits == 64) );
  ft.trackRate = sr * (double)inBits / (double)outBits;

  if ((ft.trackRate < 0.020)||(ft.trackRate > 16.0)) {
    fprintf(stderr, "%s: output trackRate outside 0.020--16.0MHz\n", argv[0]);
    exit(EXIT_FAILURE);
  }
  framesPerSec = ft.trackRate * 50.0;  /* 16MHz * 50 frames == 800 frames/s */
  ft.frameIncr = (int)(800.0 / framesPerSec);  /* +1 for 16MHz, +2 for 8MHz... */
  /* Formatter time will advance max. 1sec per one frame. */
  if ((ft.frameIncr < 1)||(ft.frameIncr > 800)) {
    fprintf(stderr, "%s: output trackRate outside 0.020--16.0MHz\n", argv[0]);
    exit(EXIT_FAILURE);
  }
  /* The smallest formatter time counter unit is 0.125msec. */
  if (((double)ft.frameIncr * framesPerSec) != 800.0) {
    fprintf(stderr, "%s: frameIncr (%f) not an integer multiple of 0.125msec\n", argv[0], (800.0 / framesPerSec));
    exit(EXIT_FAILURE);
  }
  /* Convert fractional seconds into frame counter 0--799. */
  /* (Typically ./sf is started to format data starting on integer sec.) */
  ft.f800 = 800.0*(fractSec - (double)(ft.ss));

#if 1
  fprintf(stderr, "ft = %04d-%02d-%02dT%02d:%02d:%02d f800=%d, frameIncr=%d, trackRate=%.3fMHz\n"
         "   = %04d-%03d, MJD = %d, day seconds = %d\n",
         ft.yyyy,
         ft.mm,
         ft.dd,
         ft.hh,
         ft.min,
         ft.ss,
         ft.f800,
         ft.frameIncr,
         ft.trackRate,
         ft.yyyy,
         ft.doy,
         ft.mjd,
         ft.dsssss
         );
#endif
}  /* setupFormatterStartTime */

ssize_t
sureRead(int filedes, ptByte memblk, size_t blksize)
{
  size_t nread;

  /* Read a block; if not enough, read more. */
  assert( (nread = read (filedes, memblk, blksize)) >= 0 );
  if (nread <= 0)
    goto lReturn;
  while (nread < blksize) {
    size_t nr;

    /* Not one full block in buffer, read more. */
    assert( (nr = read (filedes, memblk+nread, blksize-nread)) >= 0 );
    if (nr <= 0)
      goto lReturn;
    nread += nr;
  }  /* while not at least one full block in buffer to return */
 lReturn:
  return (nread);
}  /* sureRead */

ssize_t
sureWrite(int filedes, ptByte memblk, size_t blksize)
{
  size_t nwritten;

  /* Write a block; if couldn't write all, write again more. */
  assert( (nwritten = write (filedes, memblk, blksize)) >= 0 );
  while (nwritten < blksize) {
    size_t nw;

    /* Not everything in buffer written, write more. */
    assert( (nw = write (filedes, memblk+nwritten, blksize-nwritten)) >= 0 );
    nwritten += nw;
  }  /* while not everything in buffer written */
  return (nwritten);
}  /* sureWrite */


int
main(int argc, char *argv[])
{
  int useParity;
  size_t rdFrSize;  /* how many bytes to read frame data in */
  size_t wrFrSize;  /* how many bytes to write formatted frame data out */
  size_t dataFrSize;  /* data part of frame, in bytes */
  size_t dataFrSizeIn64;  /* data part of frame, in 64-bit words */
  ptByte wrFrame;  /* allocated buffer */
  ptByte rdFrame;  /* points to either start or in middle of wrFrame */
  ptByte crcBl;  /* allocated */
  ssize_t nread;
  ssize_t nwritten;

  int j;
  tByte i3,i2,i1,i0;

  if (argc < 8) {
    fprintf(stderr, "%s: usage: %s startTime vlba|mk4 2-bitmapping inBits inMsps outBits fanOutVariant\n"
" startTime = ISO-8601 std. UTC date/time, e.g. 2004-12-31T23:55:40\n"
" vlba|mk4 = output format, vlba non-data replacement or mk4 DR\n"
" 2-bitmapping = four digits like \"0123\", map 2-bit values 0..3 to any 0..3\n"
" inBits = how many input bit streams, 8/16/32/64\n"
" inMsps = raw (no parity) input sampling rate in decimal Msps, affects timecode\n"
" outBits = how many output bit streams, 8/16/32/64\n"
" fanOutVariant = 0, 1,..., 0==default from Mk4 memo 230rev1.2\n"
            , argv[0], argv[0]);
    exit(EXIT_FAILURE);
  }

  /* Process command line arguments. */
  setupFormatterStartTime(argv);

  /* xxx: should add to-NRZM XOR conversion to output in parity==1 mode... */
  useParity = 0;

  /* Determine mk3/vlba (=> block sizes), fan-out & crosspoint, */
  /* bit width of formatter output (=> select fmt routine variants). */

  {
    int bypw = outBits / 8;

    /* Input data for a single frame is always 2500*8 sampler data words. */
    rdFrSize = 2500*(useParity?9:8) * bypw;
    if (formatterMode == 'v') {    
      wrFrSize = (2500 + 20)*(useParity?9:8) * bypw;  /* NDR mode */
      dataFrSize = rdFrSize;  /* == 2500*(useParity?9:8) * bypw */
    } else {
      wrFrSize = rdFrSize;  /* data replacement mode */
      dataFrSize = (2500 - 20)*(useParity?9:8) * bypw;  /* DR mode */
    }
    /* Data part is bit-rearranged in 64bit words. */
    assert( (dataFrSize % 8) == 0 );
    dataFrSizeIn64 = dataFrSize / 8;

    /* Allocate frame buffer(s). */
    assert( (wrFrame = calloc(wrFrSize, 1)) != NULL );
    rdFrame = wrFrame;
    if (formatterMode == 'v') {
      /* In VLBA mode data is read into middle of write buffer. */
      /* "12+2500+8 bytes+parity" */
      rdFrame = wrFrame + (12*(useParity?9:8) * bypw);
    }
    /* Allocate largest 64 bit wide and 16 bit deep CRC buffer. */
    /* (12 words would be sufficient for Mk4.) */
    assert( (crcBl = calloc(16, sizeof(tD64))) != NULL );
  }


  if (inBits == outBits) {

  /* Init 1:1 2-bit crosspoint. (2-bit sample coding mapping is done.) */
  for (i3=0; i3<4; i3++) {
    for (i2=0; i2<4; i2++) {
      for (i1=0; i1<4; i1++) {
        for (i0=0; i0<4; i0++) {
          tByte iii = (i3<<6)|(i2<<4)|(i1<<2)|i0;

          xpt0[iii] =
            ((sm[i0]&1)    )|((sm[i0]&2)/*<<1-1*/ )|((sm[i1]&1)<<2 )|((sm[i1]&2)<<(3-1 ))
           |((sm[i2]&1)<<4 )|((sm[i2]&2)<<(5-1 ))|((sm[i3]&1)<<6 )|((sm[i3]&2)<<(7-1 ));
          xpt1[iii] = xpt0[iii]<<8;
          xpt2[iii] = xpt0[iii]<<16;
          xpt3[iii] = xpt0[iii]<<24;
#if B64
          xpt4[iii] = xpt0[iii]<<32;
          xpt5[iii] = xpt0[iii]<<40;
          xpt6[iii] = xpt0[iii]<<48;
          xpt7[iii] = xpt0[iii]<<56;
#endif
        }
      }
    }
  }
  fprintf(stderr, "%s: fan-out 1:2, 2-bit, %d -> %d bits\n", argv[0], inBits, outBits);
  }

  else if ((inBits == 16) && (outBits == 32)) {

  /* Init 1:2 2-bit crosspoint. (2x two 16-bit words to one 32-bit.) */
  for (i3=0; i3<4; i3++) {
    for (i2=0; i2<4; i2++) {
      for (i1=0; i1<4; i1++) {
        for (i0=0; i0<4; i0++) {
          tByte iii = (i3<<6)|(i2<<4)|(i1<<2)|i0;

          xpt0[iii] =
            /* sign LSB, magn next */
            ((sm[i0]&1)    )|((sm[i0]&2)<<(4-1 ))|((sm[i1]&1)<<1 )|((sm[i1]&2)<<(5-1 ))
           |((sm[i2]&1)<<8 )|((sm[i2]&2)<<(12-1))|((sm[i3]&1)<<9 )|((sm[i3]&2)<<(13-1));
          xpt1[iii] = xpt0[iii]<<16;
          xpt2[iii] = xpt0[iii]<<2;
          xpt3[iii] = xpt2[iii]<<16;
#if B64
          xpt4[iii] = xpt0[iii]<<32;
          xpt5[iii] = xpt1[iii]<<32;
          xpt6[iii] = xpt2[iii]<<32;
          xpt7[iii] = xpt3[iii]<<32;
#endif
        }
      }
    }
  }
  fprintf(stderr, "%s: fan-out 1:2, 2-bit, %d -> %d bits\n", argv[0], inBits, outBits);
  }

  else if ((inBits == 8) && (outBits == 32) && (fanOutVariant == 0)) {

  /* Init 1:4 2-bit crosspoint. (Eight 8-bit words to two 64-bit.) */
  for (i3=0; i3<4; i3++) {
    for (i2=0; i2<4; i2++) {
      for (i1=0; i1<4; i1++) {
        for (i0=0; i0<4; i0++) {
          tByte iii = (i3<<6)|(i2<<4)|(i1<<2)|i0;

          xpt0[iii] =
            /* sign LSB, magn next */
            ((sm[i0]&1)    )|((sm[i0]&2)<<(8-1 ))|((sm[i1]&1)<<1 )|((sm[i1]&2)<<(9-1 ))
           |((sm[i2]&1)<<16)|((sm[i2]&2)<<(24-1))|((sm[i3]&1)<<17)|((sm[i3]&2)<<(25-1));
          xpt1[iii] = xpt0[iii]<<2;
          xpt2[iii] = xpt0[iii]<<4;
          xpt3[iii] = xpt0[iii]<<6;
          xpt4[iii] = xpt0[iii]<<32;
          xpt5[iii] = xpt1[iii]<<32;
          xpt6[iii] = xpt2[iii]<<32;
          xpt7[iii] = xpt3[iii]<<32;
        }
      }
    }
  }
  fprintf(stderr, "%s: fan-out 1:4, 2-bit, %d -> %d bits\n", argv[0], inBits, outBits);
  }

  else if ((inBits == 8) && (outBits == 32) && (fanOutVariant == 1)) {

  /* Init 1:4 2-bit crosspoint. (Eight 8-bit words to two 64-bit.) */
  /* "Parkes variant", ch1 and ch4 as in above, ch2<-->ch3 swapped. */
  for (i3=0; i3<4; i3++) {
    for (i2=0; i2<4; i2++) {
      for (i1=0; i1<4; i1++) {
        for (i0=0; i0<4; i0++) {
          tByte iii = (i3<<6)|(i2<<4)|(i1<<2)|i0;

          xpt0[iii] =
            /* sign LSB, magn next */
            ((sm[i0]&1)    )|((sm[i0]&2)<<(8-1 ))|((sm[i1]&1)<<16)|((sm[i1]&2)<<(24-1))
           |((sm[i2]&1)<<1 )|((sm[i2]&2)<<(9-1 ))|((sm[i3]&1)<<17)|((sm[i3]&2)<<(25-1));
          xpt1[iii] = xpt0[iii]<<2;
          xpt2[iii] = xpt0[iii]<<4;
          xpt3[iii] = xpt0[iii]<<6;
          xpt4[iii] = xpt0[iii]<<32;
          xpt5[iii] = xpt1[iii]<<32;
          xpt6[iii] = xpt2[iii]<<32;
          xpt7[iii] = xpt3[iii]<<32;
        }
      }
    }
  }
  fprintf(stderr, "%s: fan-out 1:4, 2-bit, %d -> %d bits (\"Parkes\", ch2<-->ch3)\n", argv[0], inBits, outBits);
  }

  else if ((inBits == 16) && (outBits == 64) && (fanOutVariant == 0)) {

  /* Init 1:4 2-bit crosspoint. (Four 16-bit words to one 64-bit.) */
  for (i3=0; i3<4; i3++) {
    for (i2=0; i2<4; i2++) {
      for (i1=0; i1<4; i1++) {
        for (i0=0; i0<4; i0++) {
          tByte iii = (i3<<6)|(i2<<4)|(i1<<2)|i0;

          xpt0[iii] =
            /* sign LSB, magn next */
            ((sm[i0]&1)    )|((sm[i0]&2)<<(8-1 ))|((sm[i1]&1)<<1 )|((sm[i1]&2)<<(9-1 ))
           |((sm[i2]&1)<<16)|((sm[i2]&2)<<(24-1))|((sm[i3]&1)<<17)|((sm[i3]&2)<<(25-1));
          xpt1[iii] = xpt0[iii]<<32;
          xpt2[iii] = xpt0[iii]<<2;
          xpt3[iii] = xpt2[iii]<<32;
          xpt4[iii] = xpt0[iii]<<4;
          xpt5[iii] = xpt4[iii]<<32;
          xpt6[iii] = xpt0[iii]<<6;
          xpt7[iii] = xpt6[iii]<<32;
        }
      }
    }
  }
  fprintf(stderr, "%s: fan-out 1:4, 2-bit, %d -> %d bits\n", argv[0], inBits, outBits);
  }

  else if ((inBits == 16) && (outBits == 64) && (fanOutVariant == 1)) {

  /* Init 1:4 2-bit crosspoint. (Four 16-bit words to one 64-bit.) */
  /* "Parkes variant", ch1 and ch4 as in above, ch2<-->ch3 swapped. */
  for (i3=0; i3<4; i3++) {
    for (i2=0; i2<4; i2++) {
      for (i1=0; i1<4; i1++) {
        for (i0=0; i0<4; i0++) {
          tByte iii = (i3<<6)|(i2<<4)|(i1<<2)|i0;

          xpt0[iii] =
            /* sign LSB, magn next */
            ((sm[i0]&1)    )|((sm[i0]&2)<<(8-1 ))|((sm[i1]&1)<<16)|((sm[i1]&2)<<(24-1))
           |((sm[i2]&1)<<1 )|((sm[i2]&2)<<(9-1 ))|((sm[i3]&1)<<17)|((sm[i3]&2)<<(25-1));
          xpt1[iii] = xpt0[iii]<<32;
          xpt2[iii] = xpt0[iii]<<2;
          xpt3[iii] = xpt2[iii]<<32;
          xpt4[iii] = xpt0[iii]<<4;
          xpt5[iii] = xpt4[iii]<<32;
          xpt6[iii] = xpt0[iii]<<6;
          xpt7[iii] = xpt6[iii]<<32;
        }
      }
    }
  }
  fprintf(stderr, "%s: fan-out 1:4, 2-bit, %d -> %d bits (\"Parkes\", ch2<-->ch3)\n", argv[0], inBits, outBits);
  }

  else {
    fprintf(stderr, "%s: don't know yet how to map %d bits -> %d bits (fan-out variant %d)\n", argv[0], inBits, outBits, fanOutVariant);
    exit(EXIT_FAILURE);
  }

  /* Take one frame at a time from stdin, format, and write to stdout. */
  while ((nread = sureRead(STDIN_FILENO, rdFrame, rdFrSize)) > 0) {
    ptByte t = wrFrame;

    if (nread != rdFrSize) {
      fprintf(stderr, "%s: warning, nread = %d < rdFrSize = %d\n", argv[0], nread, rdFrSize);
    }
    
    /* Format this frame. */
    if (formatterMode == 'm') {
      if (outBits == 64) {
        t = (ptByte)mk4fmtc64((ptD64)t, (ptD64)crcBl, useParity);
        t = (ptByte)mk4fmtf64((ptD64)t, (ptD64)crcBl, useParity);
      } else if (outBits == 32) {
        t = (ptByte)mk4fmtc32((ptD32)t, (ptD32)crcBl, useParity);
        t = (ptByte)mk4fmtf32((ptD32)t, (ptD32)crcBl, useParity);
      } else if (outBits == 16) {
        t = (ptByte)mk4fmtc16((ptD16)t, (ptD16)crcBl, useParity);
        t = (ptByte)mk4fmtf16((ptD16)t, (ptD16)crcBl, useParity);
      } else if (outBits == 8) {
        t = (ptByte)mk4fmtc8((ptD8)t, (ptD8)crcBl, useParity);
        t = (ptByte)mk4fmtf8((ptD8)t, (ptD8)crcBl, useParity);
      } else {
        fprintf(stderr, "%s: outBits=%d, not 8/16/32/64\n", argv[0], outBits);
        exit(EXIT_FAILURE);
      }
      /* t is left at the start address of data inside frame */
    } else if (formatterMode == 'v') {
      if (outBits == 64) {
        t = (ptByte)vlbafmtc64((ptD64)t, (ptD64)crcBl, useParity);
        t = (ptByte)vlbafmtf64((ptD64)t, (ptD64)crcBl, useParity);
        assert( t == rdFrame );
        t += dataFrSize;
        t = (ptByte)vlbafmta64((ptD64)t, useParity);
      } else if (outBits == 32) {
        t = (ptByte)vlbafmtc32((ptD32)t, (ptD32)crcBl, useParity);
        t = (ptByte)vlbafmtf32((ptD32)t, (ptD32)crcBl, useParity);
        assert( t == rdFrame );
        t += dataFrSize;
        t = (ptByte)vlbafmta32((ptD32)t, useParity);
      } else if (outBits == 16) {
        t = (ptByte)vlbafmtc16((ptD16)t, (ptD16)crcBl, useParity);
        t = (ptByte)vlbafmtf16((ptD16)t, (ptD16)crcBl, useParity);
        assert( t == rdFrame );
        t += dataFrSize;
        t = (ptByte)vlbafmta16((ptD16)t, useParity);
      } else if (outBits == 8) {
        t = (ptByte)vlbafmtc8((ptD8)t, (ptD8)crcBl, useParity);
        t = (ptByte)vlbafmtf8((ptD8)t, (ptD8)crcBl, useParity);
        assert( t == rdFrame );
        t += dataFrSize;
        t = (ptByte)vlbafmta8((ptD8)t, useParity);
      } else {
        fprintf(stderr, "%s: outBits=%d, not 8/16/32/64\n", argv[0], outBits);
        exit(EXIT_FAILURE);
      }
      t = rdFrame;  /* start address of data inside frame */
    } else {
      fprintf(stderr, "%s: don't know how to format in '%c' mode\n", argv[0], formatterMode);
      exit(EXIT_FAILURE);
    }

    /* Do crosspoint switch & fan-out i.e. rearrange bits. */
    /* (Start at 't'.) */
    /* (Do always, in 1:1 mode, too, to re-code 2-bit samples. */
    /*** if (inBits != outBits) ***/ {
      for (j=0; j<(dataFrSizeIn64); j++) {
        ptByte p = t;

        tByte i0 = p[0];
        tByte i1 = p[1];
        tByte i2 = p[2];
        tByte i3 = p[3];
#if B64
        tByte i4 = p[4];
        tByte i5 = p[5];
        tByte i6 = p[6];
        tByte i7 = p[7];
#endif

        *((ptD64)t)++ = xpt0[i0] | xpt1[i1] | xpt2[i2] | xpt3[i3]
#if B64
          | xpt4[i4] | xpt5[i5] | xpt6[i6] | xpt7[i7]
#endif
          ;
      }  /* for each data word */
      if (formatterMode == 'm') {
        assert( t == (wrFrame + wrFrSize) );
      } else {
        assert( t == (rdFrame + dataFrSize) );
      }
    }  /* if bit-rearrange required */

    /* Write out formatted frame. */
    assert( ((nwritten = sureWrite(STDOUT_FILENO, wrFrame, wrFrSize)) == wrFrSize) );

    /* Advance formatter clock by 1/800sec "ticks". */
    ft.f800 += ft.frameIncr;
    if (ft.f800 > 799) {
      ft.ss += 1;
      ft.dsssss += 1;
      ft.f800 -= 800;
      if (ft.ss > 59) {
        ft.min += 1;
        ft.ss -= 60;
        if (ft.min > 59) {
          ft.hh += 1;
          ft.min -= 60;
          if (ft.hh > 23) {
            int daysInThisYear;

            /* ft.dd and ft.mm are not updated, never used in format */
            ft.doy += 1;  /* used in Mk4 format */
            ft.mjd += 1;  /* used in VLBA format */
            ft.dsssss = 0;
            ft.hh -= 24;
            daysInThisYear = 
              !(ft.yyyy % 400)||((ft.yyyy % 100)&&!(ft.yyyy % 4)) ? 366:365;
            if (ft.doy > daysInThisYear) {
              ft.yyyy += 1;
              ft.doy = 1;
            }  /* if years incremented */
          }  /* if days incremented */
        }  /* if hours incremented */
      }  /* if minutes incremented */
    }  /* if seconds incremented */

#if 0
  fprintf(stderr, "ft = %04d-%02d-%02dT%02d:%02d:%02d f800=%d, frameIncr=%d, trackRate=%.3fMHz\n"
         "   = %04d-%03d, MJD = %d, day seconds = %d\n",
         ft.yyyy,
         ft.mm,
         ft.dd,
         ft.hh,
         ft.min,
         ft.ss,
         ft.f800,
         ft.frameIncr,
         ft.trackRate,
         ft.yyyy,
         ft.doy,
         ft.mjd,
         ft.dsssss
         );
#endif

  }  /* while read something */
  if (nread < 0) {
    /* Read error, print out message and exit. */
    perror(argv[0]);
    exit(EXIT_FAILURE);
  }
  /* nread == 0, EOF, quit. */
  exit(EXIT_SUCCESS);



#if 0
  /* If ever used again, must add *NFRAMES to calloc of buffer... */

/* Testing configuration of the main(). */
#define RPTS (10)
#define NFRAMES (500)
#define XPT 1


  /* Mark4 tests. */
  {
  int i;
  int fid;
  int r;

  if (formatterMode == 'm') {

  for (r=0; r<RPTS; r++) {
    ptD32 t;

    printf("Rept %d, Start address = %08X\n", r, (unsigned int)wrFrame);
    t = (ptD32)wrFrame;
    for (i=0; i<NFRAMES; i++) {
      t = mk4fmtc32(t, (ptD32)crcBl, useParity);
#if 0
      printf("TC  address   = %08X, bits = %d\n", (unsigned int)t, t-(ptD32)wrFrame);
#endif
      t = mk4fmtf32(t, (ptD32)crcBl, useParity);
      t += 2480*(useParity?9:8);
    }  /* for NFRAMES */
    printf("End address   = %08X, bits = %d\n", (unsigned int)t, t-(ptD32)wrFrame);
  }  /* for RPTS */

  assert( (fid = creat("mk4frame", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) != -1 );
  assert( (write(fid, wrFrame, wrFrSize) == wrFrSize) );
  assert( (close(fid) != -1) );


  /* VLBA tests. */

  } else if (formatterMode == 'v') {

  for (r=0; r<RPTS; r++) {
    ptD64 t;
    int i;

#if XPT
#else
    if (r<1)
#endif
    printf("Rept %d, Start address = %08X\n", r, (unsigned int)wrFrame);
    t = (ptD64)wrFrame;
    for (i=0; i<NFRAMES; i++) {
      t = vlbafmtc64(t, (ptD64)crcBl, useParity);
#if 0
      printf("TC  address   = %08X, bits = %d\n", (unsigned int)t, t-(ptD64)wrFrame);
#endif
      t = vlbafmtf64(t, (ptD64)crcBl, useParity);
#if 0
      printf("End address   = %08X, bits = %d\n", (unsigned int)t, t-(ptD64)wrFrame);
#endif
#if XPT
      /* Up to 1:4 fan-out with full crosspoint bit-setting. */
      for (j=0; j<(2500*(useParity?9:8)); j++) {
        ptByte p = (ptByte)t;
        tByte i0 = p[0];
        tByte i1 = p[1];
        tByte i2 = p[2];
        tByte i3 = p[3];
#if B64
        tByte i4 = p[4];
        tByte i5 = p[5];
        tByte i6 = p[6];
        tByte i7 = p[7];
#endif

        *t++ = xpt0[i0] | xpt1[i1] | xpt2[i2] | xpt3[i3]
#if B64
          | xpt4[i4] | xpt5[i5] | xpt6[i6] | xpt7[i7]
#endif
          ;
      }  /* for each data word */
#else
      /* Skip data. */
      t += 2500*(useParity?9:8);
#endif
#if 0
      printf("Aux start address = %08X\n", (unsigned int)t);
#endif
      t = vlbafmta64(t, useParity);
    }  /* for NFRAMES */
#if XPT
#else
    if (r+1>=RPTS)
#endif
    printf("Aux end address   = %08X, bits = %d\n", (unsigned int)t, t-(ptD64)wrFrame);
  }  /* for repts */

  assert( (fid = creat("vlbaframe", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) != -1 );
  assert( (write(fid, wrFrame, wrFrSize)) == wrFrSize );
  assert( (close(fid) != -1) );

  }  /* if vlba */

  else {
    fprintf(stderr, "%s: don't know how to format '%c'\n", argv[0], formatterMode);
    exit(EXIT_FAILURE);
  }
  }

#endif

  exit (EXIT_SUCCESS);
}  /* main */
